/*
 * Mirai. Copyright (C) 2024 Vyn
 * This file is licensed under version 3 of the GNU General Public License (GPL-3.0-only)
 * The license can be found in the LICENSE file or at https://www.gnu.org/licenses/gpl-3.0.txt
 */

#include "TasksView.h"
#include "core/Mirai.h"
#include "utils.h"
#include <algorithm>
#include <cstddef>
#include <iostream>
#include <memory>
#include <ostream>
#include <stdexcept>
#include <string>

namespace mirai
{

TasksView::TasksView(Mirai *miraiInstance) : mirai(miraiInstance)
{
	if (!miraiInstance) {
		throw std::runtime_error("NULL pointer passed in TasksView");
	}
	update();
}

TaskItem &TasksView::operator[](int index)
{
	return *(tasksToShow.at(index));
}

size_t TasksView::count() const
{
	return tasksToShow.size();
}

void TasksView::update()
{
	tasksToShow.clear();
	for (auto &file : mirai->getFiles()) {
		if (filesFilter.size() > 0 && !vectorUtils::contains(filesFilter, file->getName())) {
			continue;
		}
		for (auto &task : file->getTasks()) {
			if (tagsFilter.size() > 0 && !vectorUtils::containsAll(tagsFilter, task->getTags())) {
				continue;
			}
			tasksToShow.push_back(task.get());
		}
	}

	std::ranges::sort(tasksToShow, [](const TaskItem *t1, const TaskItem *t2) {
		if (t1->hasDate() && !t2->hasDate()) {
			return true;
		} else if (!t1->hasDate() && t2->hasDate()) {
			return false;
		}

		if (t1->getDate() < t2->getDate()) {
			return true;
		} else if (t1->getDate() > t2->getDate()) {
			return false;
		}
		if (t1->hasTime() && !t2->hasTime()) {
			return true;
		} else if (!t1->hasTime() && t2->hasTime()) {
			return false;
		}
		return t1->getStartTime() < t2->getStartTime();
	});
}

void TasksView::addTagFilter(const std::string &tag)
{
	tagsFilter.push_back(tag);
	update();
}

void TasksView::removeTagFilter(const std::string &tag)
{
	tagsFilter
		.erase(std::remove_if(tagsFilter.begin(), tagsFilter.end(), [&](const auto &tagInFilter) {
			return tag == tagInFilter;
		}));

	update();
}

void TasksView::addFileFilter(const std::string &fileName)
{
	filesFilter.push_back(fileName);
	update();
}

void TasksView::removeFileFilter(const std::string &fileName)
{
	filesFilter.erase(std::remove_if(
		filesFilter.begin(), filesFilter.end(),
		[&](const auto &fileInFilter) {
			return fileName == fileInFilter;
		}
	));
	update();
}

void TasksView::removeFilters()
{
	tagsFilter.clear();
	update();
}

const std::vector<std::string> &TasksView::getActiveTagsFilter()
{
	return tagsFilter;
}

const std::vector<std::string> &TasksView::getActiveFilesFilter()
{
	return filesFilter;
}
} // namespace mirai
